import numpy as np
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split

from kondo_ml.instance_selection.base import SelectorMixin
from kondo_ml.instance_selection.SELCON.datasets import get_data
from kondo_ml.instance_selection.SELCON.linear import Regression


class SelconSelector(SelectorMixin, BaseEstimator):
    def __init__(self, val_frac=0.1, subsize_frac=0.8):
        super().__init__(subsize_frac)
        self.val_frac = val_frac

    def fit(self, X, y):
        indices = np.arange(X.shape[0])
        idx_train, idx_val = train_test_split(
            indices, test_size=self.val_frac, shuffle=False
        )
        self.idx_train = idx_train
        self.X_train, self.X_val, self.y_train, self.y_val = get_data(
            X[idx_train, :], X[idx_val, :], y[idx_train], y[idx_val]
        )
        self.labels = np.ones(X.shape[0], dtype="int8") * -1
        self.scores = np.ones(X.shape[0], dtype="int8") * -1
        return self

    def predict(self, X, y):
        reg = Regression()
        reg.train_model(
            self.X_train,
            self.y_train,
            self.X_val,
            self.y_val,
            fraction=self.subsize_frac,
        )
        subset_idxs = reg.return_subset()
        self.labels[self.idx_train[subset_idxs]] = 1
        self.scores = self.labels
        return self.labels



import numpy as np
from sklearn.base import BaseEstimator
from sklearn.neighbors import NearestNeighbors

from kondo_ml.instance_selection.base import SelectorMixin
from kondo_ml.utils import train_lr_model


class RegCnnSelector(SelectorMixin, BaseEstimator):
    """Adaption of the Condensed Nearest Neighbor Algorithm by (Hart 1968) for Regression. The adaptation
    was presented in the following paper: #TODO . The RegCNN algorithm removes instances that are redundant/very
    similar to instances already added to the subset.
    """

    def __init__(self, alpha=0.25, nr_of_neighbors=7, subsize_frac=1):
        super().__init__(subsize_frac=subsize_frac)
        self.k = nr_of_neighbors
        self.alpha = alpha

    def fit(self, X, y):
        """
        Fit the algorithm according to the given training data.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and
            `n_features` is the number of features.
        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self
            Fitted estimator.
        """
        self.nr_of_samples = X.shape[0]
        self.labels = np.ones(self.nr_of_samples, dtype="int8") * -1
        self.scores = np.zeros(self.nr_of_samples, dtype="float32")
        return self

    def predict(self, X, y):
        """Predict the labels (1 use for training, -1 rejected) of X according to RegCNN

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and
            `n_features` is the number of features.
        y : array-like of shape (n_samples,)
            Target vector relative to X.

        Returns
        -------
        labels: ndarray of shape (n_samples,)
            Returns +1 for samples that should be used for model training, -1 for those rejected
        """
        subset_mask = np.zeros(self.nr_of_samples, dtype="bool")  # 1
        subset_mask[0] = True  # 2
        nn_mask = np.ones(self.nr_of_samples, dtype="bool")
        for i in range(1, self.nr_of_samples):
            if nn_mask.sum() <= self.k:
                self.labels[subset_mask] = 1
                return self.labels
            investigated_instance = X[i, :].reshape(1, -1)
            model = train_lr_model(X[subset_mask, :], y[subset_mask])
            y_pred = model.predict(investigated_instance)  # 3
            # As the closest neighbor is always the instance itself, we add one neighbor and ignore the 0 index
            nbrs = NearestNeighbors(n_neighbors=self.k + 1, algorithm="auto").fit(
                X[nn_mask, :]
            )
            indices = nbrs.kneighbors(investigated_instance, return_distance=False)
            indices = indices[0, 1:]  # 5
            theta = self.alpha * np.std(y[indices])  # 6
            y_true = y[i]
            self.scores[i] = (theta - np.abs(y_true - y_pred)) * -1
            if np.abs(y_true - y_pred) > theta:  # 7
                subset_mask[i] = True  # 8
                nn_mask[i] = False  # 9
        self.labels[subset_mask] = 1
        return self.labels



import numpy as np
import torch
from sklearn import preprocessing
from sklearn.preprocessing import MinMaxScaler, StandardScaler

"""def gen_rand_prior_indices(curr_size,num_cls,x_trn,y_trn,remainList=None):

    per_sample_budget = int(curr_size/num_cls)
    if remainList is None:
        per_sample_count = [len(torch.where(y_trn == x)[0]) for x in np.arange(num_cls)]
        total_set = list(np.arange(N))
    else:
        per_sample_count = [len(torch.where(y_trn[remainList] == x)[0]) for x in np.arange(num_cls)]
        total_set = remainList
    indices = []
    count = 0
    for i in range(num_cls):
        if remainList is None:
            label_idxs = torch.where(y_trn == i)[0].cpu().numpy()
        else:
            label_idxs = torch.where(y_trn[remainList] == i)[0].cpu().numpy()
            label_idxs = np.array(remainList)[label_idxs]

        if per_sample_count[i] > per_sample_budget:
            indices.extend(list(np.random.choice(label_idxs, size=per_sample_budget, replace=False)))
        else:
            indices.extend(label_idxs)
            count += (per_sample_budget - per_sample_count[i])
    for i in indices:
        total_set.remove(i)
    indices.extend(list(np.random.choice(total_set, size= count, replace=False)))
    return indices"""


def get_slices(data_name, data, labels, device, buckets=None, clean=True):
    # data_slices = []
    # abel_slices =[]

    val_data_slices = []
    val_label_slices = []

    tst_data_slices = []
    tst_label_slices = []

    """sc = StandardScaler()
    x_trn = sc.fit_transform(x_trn)
    x_val = sc.transform(x_val)
    x_tst = sc.transform(x_tst)"""

    if data_name == "Community_Crime_old":
        protect_feature = [2, 3, 4, 5]

        data_class = []

        N = int(0.1 * len(data) / (buckets * len(protect_feature)))

        total_set = set(list(np.arange(len(data))))

        for i in protect_feature:
            digit = np.ones(data.shape[0], dtype=np.int8) * (i - 1)
            low = np.min(data[:, i])
            high = np.max(data[:, i])
            bins = np.linspace(low, high, buckets)
            digitized = np.digitize(data[:, i], bins)
            digit = digit * 10 + digitized

            classes, times = np.unique(digit, return_counts=True)
            times, classes = zip(*sorted(zip(times, classes)))
            data_class.append(classes)

            count = 0
            for cl in classes[:-1]:
                indices = []
                indices_tst = []

                idx = (digit == cl).nonzero()[0].flatten()
                idx.tolist()
                idxs = set(idx)
                idxs.intersection_update(total_set)
                idx = list(idxs)
                # print(cl,len(idx))

                curr_N = int(len(idx) / 3)

                # print(curr_N,N)

                indices.extend(
                    list(np.random.choice(idx, size=min(N, curr_N), replace=False))
                )
                total_set.difference(indices)
                idxs.difference(indices)
                idx = list(idxs)
                indices_tst.extend(
                    list(np.random.choice(idx, size=min(N, curr_N), replace=False))
                )
                total_set.difference(indices_tst)

                if curr_N < N:
                    count += N - curr_N

                val_data_slices.append(
                    torch.from_numpy(data[indices]).float().to(device)
                )
                val_label_slices.append(
                    torch.from_numpy(labels[indices]).float().to(device)
                )

                tst_data_slices.append(
                    torch.from_numpy(data[indices_tst]).float().to(device)
                )
                tst_label_slices.append(
                    torch.from_numpy(labels[indices_tst]).float().to(device)
                )

            indices = []
            indices_tst = []

            idx = (digit == classes[-1]).nonzero()[0].flatten()
            idx.tolist()
            idxs = set(idx)
            idxs.intersection_update(total_set)
            idx = list(idxs)

            indices.extend(list(np.random.choice(idx, size=N + count, replace=False)))
            total_set.difference(indices)
            idxs.difference(indices)
            idx = list(idxs)
            indices_tst.extend(
                list(np.random.choice(idx, size=N + count, replace=False))
            )
            total_set.difference(indices_tst)

            val_data_slices.append(torch.from_numpy(data[indices]).float().to(device))
            val_label_slices.append(
                torch.from_numpy(labels[indices]).float().to(device)
            )

            tst_data_slices.append(
                torch.from_numpy(data[indices_tst]).float().to(device)
            )
            tst_label_slices.append(
                torch.from_numpy(labels[indices_tst]).float().to(device)
            )

        final_lables = [j for sub in data_class for j in sub]
        left = list(total_set)
        data_left = data[left]
        label_left = labels[left]

    elif data_name == "OnlineNewsPopularity":
        protect_feature = [11, 12, 13, 14, 15, 16]

        final_lables = [
            "Lifestyle",
            "Entertainment",
            "Business",
            "Social Media",
            "Tech",
            "World",
        ]

        total_set = set(list(np.arange(len(data))))

        N = int(0.1 * len(data) / len(protect_feature))

        max_times = 0

        for pf in protect_feature:
            classes, times = np.unique(data[:, pf], return_counts=True)

            one_id = (classes == 1.0).nonzero()[0].flatten()[0]

            if max_times < times[one_id]:
                max_times = times[one_id]
                max_id = pf

        most = final_lables[max_id - protect_feature[0]]
        final_lables.remove(most)
        final_lables.append(most)

        count = 0

        for pf in protect_feature:
            if pf == max_id:
                continue

            idx = (data[:, pf] == 1.0).nonzero()[0].flatten()
            idx.tolist()
            idxs = set(idx)
            idxs.intersection_update(total_set)
            idx = list(idxs)
            # print(cl,len(idx))

            curr_N = int(len(idx) / 3)

            # print(curr_N,N)

            indices = list(np.random.choice(idx, size=min(N, curr_N), replace=False))
            total_set.difference(indices)
            idxs.difference(indices)
            idx = list(idxs)
            indices_tst = list(
                np.random.choice(idx, size=min(N, curr_N), replace=False)
            )
            total_set.difference(indices_tst)

            if curr_N < N:
                count += N - curr_N

            """val_data_slices.append(torch.from_numpy(data[indices]).float().to(device))
            val_label_slices.append(torch.from_numpy(labels[indices]).float().to(device))

            tst_data_slices.append(torch.from_numpy(data[indices_tst]).float().to(device))
            tst_label_slices.append(torch.from_numpy(labels[indices_tst]).float().to(device))"""

            val_data_slices.append(data[indices])
            val_label_slices.append(labels[indices])

            tst_data_slices.append(data[indices_tst])
            tst_label_slices.append(labels[indices_tst])

        idx = (data[:, max_id] == 1.0).nonzero()[0].flatten()
        idx.tolist()
        idxs = set(idx)
        idxs.intersection_update(total_set)
        idx = list(idxs)

        indices = list(np.random.choice(idx, size=N + count, replace=False))
        total_set.difference(indices)
        idxs.difference(indices)
        idx = list(idxs)
        indices_tst = list(np.random.choice(idx, size=N + count, replace=False))
        total_set.difference(indices_tst)

        """val_data_slices.append(torch.from_numpy(data[indices]).float().to(device))
        val_label_slices.append(torch.from_numpy(labels[indices]).float().to(device))

        tst_data_slices.append(torch.from_numpy(data[indices_tst]).float().to(device))
        tst_label_slices.append(torch.from_numpy(labels[indices_tst]).float().to(device))"""

        val_data_slices.append(data[indices])
        val_label_slices.append(labels[indices])

        tst_data_slices.append(data[indices_tst])
        tst_label_slices.append(labels[indices_tst])

        left = list(total_set)
        sc = MinMaxScaler()  # StandardScaler()
        sc_l = MinMaxScaler()

        # print(data[left][0])
        data_left = sc.fit_transform(data[left])
        label_left = np.reshape(
            sc_l.fit_transform(np.reshape(labels[left], (-1, 1))), (-1)
        )
        # print(data_left[0])
        # preprocessing.normalize(data[left])

        for j in range(len(val_data_slices)):
            val_data_slices[j] = (
                torch.from_numpy(sc.transform(val_data_slices[j])).float().to(device)
            )
            tst_data_slices[j] = (
                torch.from_numpy(sc.transform(tst_data_slices[j])).float().to(device)
            )

            val_label_slices[j] = (
                torch.from_numpy(
                    np.reshape(
                        sc_l.transform(np.reshape(val_label_slices[j], (-1, 1))), (-1)
                    )
                )
                .float()
                .to(device)
            )
            tst_label_slices[j] = (
                torch.from_numpy(
                    np.reshape(
                        sc_l.transform(np.reshape(tst_label_slices[j], (-1, 1))), (-1)
                    )
                )
                .float()
                .to(device)
            )

    elif data_name in ["census", "LawSchool", "German_credit", "Community_Crime"]:
        if data_name == "census":
            protect_feature = 8  # 9
        elif data_name == "LawSchool":
            protect_feature = 0
        elif data_name == "German_credit":
            protect_feature = 8
        elif data_name == "Community_Crime":
            protect_feature = -1

        total_set = set(list(np.arange(len(data))))

        classes, times = np.unique(data[:, protect_feature], return_counts=True)
        times, classes = zip(*sorted(zip(times, classes)))

        # print(times)
        # print(classes)

        N = int(0.1 * len(data) / len(classes))

        count = 0
        for cl in classes[:-1]:
            indices = []
            indices_tst = []

            idx = (data[:, protect_feature] == cl).nonzero()[0].flatten()
            idx.tolist()

            curr_N = int(len(idx) / 3)

            # print(curr_N,N)

            indices.extend(
                list(np.random.choice(idx, size=min(N, curr_N), replace=False))
            )
            total_set.difference(indices)
            idxs = set(idx)
            idxs.difference(indices)
            idx = list(idxs)
            indices_tst.extend(
                list(np.random.choice(idx, size=min(N, curr_N), replace=False))
            )
            total_set.difference(indices_tst)

            if curr_N < N:
                count += N - curr_N

            # val_data_slices.append(torch.from_numpy(preprocessing.normalize(data[indices])).float().to(device))
            val_data_slices.append(torch.from_numpy(data[indices]).float().to(device))
            val_label_slices.append(
                torch.from_numpy(labels[indices]).float().to(device)
            )

            # tst_data_slices.append(torch.from_numpy(preprocessing.normalize(data[indices_tst])).float().to(device))
            tst_data_slices.append(
                torch.from_numpy(data[indices_tst]).float().to(device)
            )
            tst_label_slices.append(
                torch.from_numpy(labels[indices_tst]).float().to(device)
            )

        indices = []
        indices_tst = []

        idx = (data[:, protect_feature] == classes[-1]).nonzero()[0].flatten()
        idx.tolist()

        indices.extend(list(np.random.choice(idx, size=N + count, replace=False)))
        total_set.difference(indices)
        idxs = set(idx)
        idxs.difference(indices)
        idx = list(idxs)
        indices_tst.extend(list(np.random.choice(idx, size=N + count, replace=False)))
        total_set.difference(indices_tst)

        # val_data_slices.append(torch.from_numpy(preprocessing.normalize(data[indices])).float().to(device))
        val_data_slices.append(torch.from_numpy(data[indices]).float().to(device))
        val_label_slices.append(torch.from_numpy(labels[indices]).float().to(device))

        # tst_data_slices.append(torch.from_numpy(preprocessing.normalize(data[indices_tst])).float().to(device))
        tst_data_slices.append(torch.from_numpy(data[indices_tst]).float().to(device))
        tst_label_slices.append(
            torch.from_numpy(labels[indices_tst]).float().to(device)
        )

        final_lables = classes
        left = list(total_set)
        data_left = data[left]  # preprocessing.normalize(data[left])
        label_left = labels[left]

        if not clean:
            noise_size = int(len(label_left) * 0.5)
            noise_indices = np.random.choice(
                np.arange(len(label_left)), size=noise_size, replace=False
            )

            sigma = 40
            label_left[noise_indices] = label_left[noise_indices] + np.random.normal(
                0, sigma, noise_size
            )

    return (
        data_left,
        label_left,
        val_data_slices,
        val_label_slices,
        final_lables,
        tst_data_slices,
        tst_label_slices,
        final_lables,
    )



from kondo_ml.kondo_ml import KondoMLPackage

__author__ = "Lukas Rücker"
__email__ = "ruecker.lukas@gmail.com"
__version__ = "0.0.9"

# module level doc-string
__doc__ = """
Kondo-ML
================

Description
-----------
Kondo-ML is a Python package to gather instance selection algorithms for regression problems.
The package was created as a byproduct to my master's thesis. The name Kondo is referring to 
Marie Kondo. The idea is that with the instance selection only those instances are kept
that spark joy in the model (the user).


"""



"""This is the kondo_ml.py script."""


class KondoMLPackage:
    """Kondo-ML class."""

    def __init__(self, message=None):
        """Initialize with user-defined parameters."""
        self.message = message

    def show(self):
        """Run the show function."""
        if self.message is None:
            print("No message was given as input during initialization.")
        else:
            print(self.message)

        # Return
        return None



import torch
import torch.nn as nn


class RegressionNet(nn.Module):
    def __init__(self, input_dim):
        super(RegressionNet, self).__init__()
        self.linear = nn.Linear(input_dim, 1)  # ,bias=False)
        self.feature_dim = input_dim

    def forward(self, x, last=False):
        scores = self.linear(x)
        # scores = torch.sigmoid(self.linear(x))
        # return scores.view(-1)
        if last:
            return scores.view(-1), x
        else:
            return scores.view(-1)

    def get_embedding_dim(self):
        return self.feature_dim


class LogisticNet(nn.Module):
    def __init__(self, input_dim):  # ,num_cls):
        super(LogisticNet, self).__init__()
        self.linear = nn.Linear(input_dim, 1)
        # self.feature_dim = input_dim

    def forward(self, x, last=False):
        scores = self.linear(x)
        scores = torch.sigmoid(self.linear(x).view(-1))
        return scores


class DualNet(nn.Module):
    def __init__(self, input_dim):
        super(DualNet, self).__init__()
        self.linear = nn.Linear(input_dim, 1, bias=False)

    def forward(self, x):
        scores = self.linear(x)
        return scores.view(-1)



import numpy as np
from sklearn.base import BaseEstimator

from kondo_ml.instance_selection.base import SelectorMixin


class RandomSelector(SelectorMixin, BaseEstimator):
    def __init__(self, subsize_frac):
        super().__init__(subsize_frac)

    def fit(self, X, y=None):
        self.nr_samples = X.shape[0]
        self.subsize_int = self.calc_subset_sizeint(self.nr_samples)
        self.labels = np.ones(self.nr_samples, dtype="int8") * -1
        self.scores = np.random.uniform(-1, 1, self.nr_samples)
        return self

    def predict(self, X, y=None):
        idx_sorted = np.argsort(self.scores)
        idx_selected = idx_sorted[-self.subsize_int :]
        self.labels[idx_selected] = 1
        return self.labels



import os

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler


def process_time_series(datadir, past_length, col_name, name, save_data=True):
    x_trn = []
    y_trn = []

    path = os.path.join(datadir, "prices-split-adjusted.csv")

    prices_split_adjusted = pd.read_csv(path, index_col="date", parse_dates=["date"])

    # Removing data with less than 1500 entries
    symbol = pd.DataFrame(prices_split_adjusted["symbol"].value_counts())
    omit = symbol[symbol["symbol"] < 1500].index
    prices_split_adjusted = prices_split_adjusted[
        ~prices_split_adjusted["symbol"].isin(omit)
    ]

    symbol = pd.DataFrame(prices_split_adjusted["symbol"].value_counts())

    for sym in symbol.index:
        stock = prices_split_adjusted[prices_split_adjusted["symbol"] == sym]

        data_len = len(stock.index)

        x_trn_slice = np.zeros((data_len - past_length, past_length))
        y_trn_slice = np.zeros(data_len - past_length)

        for ind in range(data_len - past_length):
            x_trn_slice[ind] = stock[col_name][stock.index[ind : ind + past_length]]
            y_trn_slice[ind] = stock[col_name][stock.index[ind + past_length]]

        x_trn.append(x_trn_slice)
        y_trn.append(y_trn_slice)

    finaL_x_trn = np.concatenate(x_trn, axis=0)
    finaL_y_trn = np.concatenate(y_trn, axis=0)

    if save_data:
        # Save the numpy files to the folder where they come from
        data_np_path = os.path.join(
            datadir, name + "_" + str(past_length) + ".data.npy"
        )
        target_np_path = os.path.join(
            datadir, name + "_" + str(past_length) + ".label.npy"
        )
        np.save(data_np_path, finaL_x_trn)
        np.save(target_np_path, finaL_y_trn)

    return finaL_x_trn, finaL_y_trn


def load_time_series_data(datadir, dset_name, past_length, clean=True):
    if os.path.isfile(
        os.path.join(datadir, dset_name + "_" + str(past_length) + ".data.npy")
    ) and os.path.isfile(
        os.path.join(datadir, dset_name + "_" + str(past_length) + ".label.npy")
    ):
        x_trn = np.load(
            os.path.join(datadir, dset_name + "_" + str(past_length) + ".data.npy")
        )
        y_trn = np.load(
            os.path.join(datadir, dset_name + "_" + str(past_length) + ".label.npy")
        )

    else:
        if dset_name == "NY_Stock_exchange_close":
            x_trn, y_trn = process_time_series(datadir, past_length, "close", dset_name)

        elif dset_name == "NY_Stock_exchange_open":
            x_trn, y_trn = process_time_series(datadir, past_length, "open", dset_name)

        elif dset_name == "NY_Stock_exchange_high":
            x_trn, y_trn = process_time_series(datadir, past_length, "high", dset_name)

        elif dset_name == "NY_Stock_exchange_low":
            x_trn, y_trn = process_time_series(datadir, past_length, "low", dset_name)

    y_trn = y_trn  # + 1600

    x_trn, x_tst, y_trn, y_tst = train_test_split(
        x_trn, y_trn, test_size=0.1, random_state=42
    )
    x_trn, x_val, y_trn, y_val = train_test_split(
        x_trn, y_trn, test_size=0.005, random_state=42
    )

    if not clean:
        noise_size = int(len(y_trn) * 0.5)
        noise_indices = np.random.choice(
            np.arange(len(y_trn)), size=noise_size, replace=False
        )

        sigma = 40
        y_trn[noise_indices] = y_trn[noise_indices] + np.random.normal(
            0, sigma, noise_size
        )

    sc = MinMaxScaler()  # StandardScaler()
    x_trn = sc.fit_transform(x_trn)
    x_val = sc.transform(x_val)
    x_tst = sc.transform(x_tst)

    """sc_l = MinMaxScaler() #StandardScaler()
    y_trn = np.reshape(sc_l.fit_transform(np.reshape(y_trn,(-1,1))),(-1))"""
    # y_val = np.reshape(sc_l.fit_transform(np.reshape(y_val,(-1,1))),(-1))
    # y_tst = np.reshape(sc_l.fit_transform(np.reshape(y_tst,(-1,1))),(-1))

    fullset = (x_trn, y_trn)
    valset = (x_val, y_val)
    testset = (x_tst, y_tst)

    return fullset, valset, testset  # , sc_l
